﻿Imports System.Net.Sockets
Imports System.Net
Imports System.Text
Imports System.Threading
Imports System.Text.RegularExpressions


Public Class frmMain

    ' Thread Receiver Declarations
    Private udpClient As New UdpClient
    Private receivingUdpClient As UdpClient
    Private RemoteIpEndPoint As New IPEndPoint(System.Net.IPAddress.Any, 0)
    Private ThreadReceive As Thread
    Private _receiverEnabled As Boolean
    ' Declare the Delegates (used to be called from external Threads)
    Private Delegate Sub processMessageDelegate(ByVal txt As String)

    ' Declare collections of the input controls.
    Private checkboxes As CheckBox()
    Private inputs As TextBox()

    ' Define the K8055 card Object
    Private card As New K8055()

    ' Default registry storage location
    Private BASE_REGISTRY_KEY As String = "HKEY_CURRENT_USER\Software\Bvsystems\K8055 BroadCaster"


    ' Enumerations
    Private Enum userMode
        SLAVE = 0
        MASTER = 1
    End Enum



#Region "Main Form"
    ''' <summary>
    ''' Form load event. The first procedure of the program.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        checkboxes = New CheckBox() {OUT1, OUT2, OUT3, OUT4, OUT5, OUT6, OUT7, OUT8}
        inputs = New TextBox() {inpBroadCastAddress, inpPort, inpKey}

        card.connect(True, True)
        If card.isConnected Then
            addLog("card is connected: Client Mode")
            status.Text = "K8055 Detected. The program starts in client mode"
            chkIamMaster.Checked = False
        Else
            addLog("card is not connected: Master Mode")
            status.Text = "No K8055 Detected. The program starts in master mode"
            chkIamMaster.Checked = True
        End If

        readFromRegistry()

        For Each check As CheckBox In checkboxes
            AddHandler check.CheckStateChanged, AddressOf CheckedChanged
        Next

        For Each input As TextBox In inputs
            AddHandler input.TextChanged, AddressOf textBoxChanged
        Next

        setUserMode(userMode.SLAVE)
        chkAutoSend.Enabled = False
        chkMasterSupportsReceive.Enabled = False
        runReceiver()
    End Sub


    ''' <summary>
    ''' This event is raised when the master checkbox is changed. It will set the program to client mode or server mode.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub chkIamMaster_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles chkIamMaster.CheckedChanged
        If chkIamMaster.Checked Then
            ' Master Mode Enabled
            setUserMode(userMode.MASTER)
            addLog("Mastermode enabled")
            chkAutoSend.Enabled = True
            chkMasterSupportsReceive.Enabled = True

            If chkMasterSupportsReceive.Checked Then
                runReceiver()
            Else
                stopReceiver()
            End If

        Else
            ' Slave Mode Enabled 
            setUserMode(userMode.SLAVE)
            addLog("Mastermode disabled. Setting 00000000")
            chkAutoSend.Enabled = False
            chkMasterSupportsReceive.Enabled = False
            runReceiver()
        End If
    End Sub


    ''' <summary>
    ''' This event is raised when the chkMasterSupportsReceive checkbox is clicked.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub chkMasterSupportsReceive_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles chkMasterSupportsReceive.CheckedChanged
        If chkIamMaster.Checked = True Then
            If chkMasterSupportsReceive.Checked Then
                runReceiver()
            Else
                stopReceiver()
            End If
        End If
    End Sub


    ''' <summary>
    ''' cmdSendToSlaves will initiate a broadcast.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub cmdSendToSlaves_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdSendToSlaves.Click
        If fieldsAreValid() Then
            addLog("broadcasting ..." & cChannelsToString())
            sendMessage()
        Else
            MsgBox("One or more input fields contain an error. Correct this first!", _
                    MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, _
                    "Data Error")
        End If
    End Sub


    ''' <summary>
    ''' setUserMode will set the current usermode for the program: will it behave as a client, or
    ''' as a server?
    ''' </summary>
    ''' <param name="mode">A mode of Enum type userMode indicating the desired mode.</param>
    ''' <remarks></remarks>
    Private Sub setUserMode(ByVal mode As userMode)
        Select Case mode
            Case userMode.MASTER
                cmdSendToSlaves.Enabled = True
            Case userMode.SLAVE
                cmdSendToSlaves.Enabled = False
        End Select

        cmdSendToSlaves.Enabled = (mode = userMode.MASTER)
        For Each check As CheckBox In checkboxes
            'If mode = userMode.MASTER Then
            check.Checked = False
            'End If

            check.Enabled = (mode = userMode.MASTER)
        Next
    End Sub


    ''' <summary>
    ''' addLog will add a new line to the output log. It includes a timestamp, followed by a text.
    ''' </summary>
    ''' <param name="newLine">the text of the line to add (without timestamp)</param>
    ''' <remarks></remarks>
    Private Sub addLog(ByVal newLine As String)

        lstLog.Items.Add(Format(Now(), "dd/MM HH:mm") & "   " & newLine)
        lstLog.SelectedIndex = lstLog.Items.Count() - 1
    End Sub


    ''' <summary>
    ''' This is an event handler which will be called when one of the INPUT checkboxes is clicked.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs)
        If chkAutoSend.Checked And chkIamMaster.Checked Then
            ' Send
            cmdSendToSlaves_Click(Nothing, Nothing)
        End If
    End Sub
#End Region


#Region "Communication and Decoding"
    ''' <summary>
    ''' stopReceiver will force the UDP receiver to stop.
    ''' For now this might raise errors.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub stopReceiver()
        addLog("UDP Receiver Stopped")
        Try
            _receiverEnabled = False
            receivingUdpClient.Close()
            ThreadReceive.Abort()
        Catch ex As Exception
            ' catch the errors. Just show them into the Debug window
            Debug.Print("Error: " & ex.StackTrace)
        End Try
    End Sub


    ''' <summary>
    ''' runReceiver will start the UDP listener. 
    ''' It created an receivingUdpClient object which listens to the requested port. The object will
    ''' be checked in a separate thread (See the receiveMessages() sub).
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub runReceiver()
        addLog("UDP Receiver Enabled")
        If Not _receiverEnabled Then
            Try
                _receiverEnabled = True

                Try
                    receivingUdpClient = New UdpClient(getPort())                                   ' Create an receiveUDPclient object
                Catch ex As Exception
                    MsgBox("Server cannot start: " & ex.Message, MsgBoxStyle.OkOnly + MsgBoxStyle.Critical)
                    Exit Sub
                End Try

                ThreadReceive = New Thread(AddressOf receiveMessages)                           ' Receive 
                ThreadReceive.IsBackground = True                                               ' Set thread properties
                ThreadReceive.Start()
            Catch ex As Exception
                ' catch the errors. Just show them into the Debug window
                Debug.Print("Error: " & ex.StackTrace)
            End Try
        End If
    End Sub


    ''' <summary>
    ''' the sub receivemessages will be executed in a separate thread. It listens for incoming data. 
    ''' The receiver is restarted each time some data was received.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub receiveMessages()
        Try
            Do
                Dim receiveBytes As [Byte]() = receivingUdpClient.Receive(RemoteIpEndPoint)     ' get the data
                Dim BitDet As BitArray = New BitArray(receiveBytes)                             ' put it into a BitArray object
                Dim returnData As String = Encoding.ASCII.GetString(receiveBytes)               ' Encode using ACII
                'receivingUdpClient.Close()                                                      ' Close the UDP Client
                Me.Invoke(New processMessageDelegate(AddressOf receivedMessage), returnData)    ' call the receivedMessage sub in the main thread

                'If _receiverEnabled Then
                '    runReceiver()                                                               ' restart the receiver (if allowed)
                'End If
            Loop
        Catch ex As Exception
            ' catch the errors. Just show them into the Debug window
            Debug.Print("Error: " & ex.StackTrace)
        End Try
    End Sub


    ''' <summary>
    ''' This delegate method will be executed when some new data was received by the receiver thread. 
    ''' </summary>
    ''' <param name="txt">The received data (encrypted)</param>
    ''' <remarks></remarks>
    Private Sub receivedMessage(ByVal txt As String)

        Dim s As New Encrypter(inpKey.Text)

        Try
            txt = s.deCrypt(txt)
        Catch ex As Exception
            txt = ""
        Finally
            Dim parameters() As String
            parameters = txt.Split(" ")

            If UBound(parameters) = 2 Then
                ' We have the right number of parameters
                If parameters(0) = "SEND" Then
                    ' The String looks okay
                    Dim r As New Regex("^[0-1]{8}$")
                    If r.IsMatch(parameters(2)) Then
                        addLog("received from " & parameters(1) & ":" & parameters(2))
                        decodeAndSendToK8055(parameters(2))
                    Else
                        addLog("Received invalid data (check password)")
                    End If
                Else
                    addLog("Received invalid data (check password)")
                End If
            Else
                addLog("Received invalid data (check password)")
            End If
        End Try
    End Sub


    ''' <summary>
    ''' decodeAndSendToK8055 decodes a String back to the K8055 channels. It will change the K8055 channels
    ''' in accordance with your desired settings.
    ''' </summary>
    ''' <param name="channels">a 8-character String containing the received channel settings.</param>
    ''' <remarks></remarks>
    Private Sub decodeAndSendToK8055(ByVal channels As String)
        For i = 0 To 7
            Dim state As String = channels.Chars(i)
            Select Case state
                Case "1"
                    card.setDigitalOutput(i + 1)
                Case "0"
                    card.clearDigitalOutput(i + 1)
            End Select
        Next
    End Sub


    ''' <summary>
    ''' getPort returns the current port
    ''' </summary>
    ''' <returns>An Integer indicating the port which was typed into the inputbox.</returns>
    ''' <remarks></remarks>
    Private Function getPort() As Integer
        Dim port As Integer = 1000 'Use port 1000 in case of an error
        If Integer.TryParse(inpPort.Text, port) Then
            Return port
        End If
    End Function


    ''' <summary>
    ''' getcheckers will return a String with 8 characters, containing only 1 or 0's. It represents the desired input state.
    ''' </summary>
    ''' <returns>A 8-character string which represents the desired output state.</returns>
    ''' <remarks></remarks>
    Private Function cChannelsToString() As String
        Dim response As String = ""
        For Each check As CheckBox In checkboxes
            response += IIf(check.Checked, "1", "0")
        Next
        Return response
    End Function


    ''' <summary>
    ''' sendMessage will send a broadcast by using an encrypted broadcast message.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub sendMessage()
        Try
            Dim bytCommand As Byte() = New Byte() {}
            Dim pret As Integer
            udpClient.Connect(inpBroadCastAddress.Text, getPort())

            Dim s As New Encrypter(inpKey.Text)
            bytCommand = Encoding.ASCII.GetBytes(s.enCrypt("SEND " & My.Computer.Name & " " & cChannelsToString()))
            pret = udpClient.Send(bytCommand, bytCommand.Length)

        Catch ex As Exception
            MsgBox("error")
        End Try
    End Sub
#End Region


#Region "Input Fields and Validation"
    ''' <summary>
    ''' fieldsAreValid is called to determine if all of the values entered in the input boxes are valid.
    ''' </summary>
    ''' <returns>Returns a boolean to indicate the inputs are valid (True means: everything is valid)</returns>
    ''' <remarks></remarks>
    Private Function fieldsAreValid() As Boolean
        Dim response = True

        Dim r As New Regex("")

        r = New Regex("^\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\b$")
        If Not r.IsMatch(inpBroadCastAddress.Text) Then
            ErrorProvider1.SetError(inpBroadCastAddress, "The broadcast is not valid. Select a valid one. For Example: 255.255.255.0")
            response = False
        Else
            ErrorProvider1.SetError(inpBroadCastAddress, "")
        End If


        Dim iPort As Integer
        If Integer.TryParse(inpPort.Text, iPort) Then
            If iPort > 0 And iPort < 65535 Then
                ErrorProvider1.SetError(inpPort, "")
            Else
                ErrorProvider1.SetError(inpPort, "Select an UDP port between 1 and 65535. Server and client must share the same ports.")
                response = False
            End If
        Else
            ErrorProvider1.SetError(inpPort, "Select an UDP port between 1 and 65535. Server and client must share the same ports.")
            response = False
        End If


        r = New Regex("^[a-zA-Z0-9]{8}$")
        If Not r.IsMatch(inpKey.Text) Then
            ErrorProvider1.SetError(inpKey, "Type a password to encrypt your communications. Password size " & vbCrLf & _
                                             "must be 8 characters exactly, containing letters and numbers " & vbCrLf & _
                                             "only. You need set the same password on all programs.")
            response = False
        Else
            ErrorProvider1.SetError(inpKey, "")
        End If

        If response = False Then
            chkAutoSend.Checked = False
            chkAutoSend.Enabled = False
        Else
            If chkIamMaster.Checked Then chkAutoSend.Enabled = True
        End If

        Return response
    End Function


    ''' <summary>
    ''' textBoxChanged is an eventhandler which is raised when one of the textboxes is changed. Then the data will
    ''' automatic be saved into the Registry.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub textBoxChanged(ByVal sender As System.Object, ByVal e As System.EventArgs)
        If fieldsAreValid() Then
            'Write to the registry, but only if the fields are valid.
            saveSettingsToRegistry()
        End If
    End Sub


    ''' <summary>
    ''' This event will check if the server has to be restarted when the remote port changed.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub inpPort_TextChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles inpPort.TextChanged
        If Not Me.Visible Then Exit Sub 'Temporary bugfix. Todo: OPTIMIZE!
        If fieldsAreValid() Then
            If chkIamMaster.Checked = False Or chkMasterSupportsReceive.Checked = True Then
                stopReceiver()
                runReceiver()
            End If
        End If
    End Sub
#End Region


#Region "Read and Write settings"
    ''' <summary>
    ''' saveSettingsToRegistry will write all settings to the Registry.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub saveSettingsToRegistry()
        For Each inp As TextBox In inputs
            My.Computer.Registry.SetValue(BASE_REGISTRY_KEY, inp.Name, inp.Text)
        Next
    End Sub

    ''' <summary>
    ''' readFromRegistry reads all settings from the Registry.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub readFromRegistry()
        For Each inp As TextBox In inputs
            inp.Text = My.Computer.Registry.GetValue(BASE_REGISTRY_KEY, inp.Name, "")
        Next
    End Sub
#End Region


End Class